
/* nbody.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nbody.h"

const float default_domain_size_x             = 1.0e+18; /* m  */
const float default_domain_size_y             = 1.0e+18; /* m  */
const float default_domain_size_z             = 1.0e+18; /* m  */
const float default_mass_maximum              = 1.0e+18; /* kg */
const float default_time_interval             = 1.0e+18; /* s  */
const int   default_number_of_particles       = 1000;
const int   default_number_of_timesteps       = 100;
const int   default_timesteps_between_outputs = 100;
const int   default_random_seed               = 12345;
const char  default_base_filename[]           = "nbody";
const char  default_arguments_filename[]      = "nbody_input.txt";

static float mass_maximum;

static const int    default_string_length = 1023;
static float        domain_size_x;
static float        domain_size_y;
static float        domain_size_z;
static float        time_interval;
static int          number_of_particles;
static int          number_of_timesteps;
static int          timesteps_between_outputs;
static unsigned int random_seed;

static void input_arguments(char* arguments_filename, char* base_filename);

int main (int argc, char** argv)
{ /* main */
    char arguments_filename[default_string_length + 1];
    char base_filename[default_string_length + 1];
    ParticleArray particle_array    = (ParticleArray)NULL;
    ForceArray    total_force_array = (ForceArray)NULL;
    int timestep;

   /*
    * Determine the name of the file of arguments to input.
    */
    strcpy(arguments_filename, default_arguments_filename);

   /*
    * Input the file of arguments.
    */
    input_arguments(arguments_filename, base_filename);

   /*
    * Construct an array of particles.
    */
    particle_array = ParticleArray_construct(number_of_particles);

   /*
    * Initialize the array of particles randomly.
    */
    ParticleArray_initialize(particle_array, number_of_particles,
                             domain_size_x, domain_size_y, domain_size_z,
                             mass_maximum);

   /*
    * Construct an array of forces.
    */
    total_force_array = ForceArray_construct(number_of_particles);

   /*
    * Output the initial array of particles.
    */
    ParticleArray_output(
        base_filename,
        particle_array,
        number_of_particles,
        0,
        number_of_timesteps,
        timesteps_between_outputs,
        domain_size_x,
        domain_size_y,
        domain_size_z,
        mass_maximum,
        time_interval,
        random_seed);

   /*
    * Loop over the timesteps.
    */
    for (timestep = 1; timestep <= number_of_timesteps; timestep++) {

       /*
        * Calculate the total force applied to each particle.
        */
        ForceArray_calculate(total_force_array,
                             particle_array, number_of_particles);

       /*
        * Move each particle based on the total force applied to it.
        */
        ParticleArray_move(particle_array, total_force_array,
                           number_of_particles, time_interval);

       /*
        * Every so often, output the list of particles.
        */
        if ((timestep % timesteps_between_outputs) == 0) {

            ParticleArray_output(
                base_filename,
                particle_array,
                number_of_particles,
                timestep,
                number_of_timesteps,
                timesteps_between_outputs,
                domain_size_x,
                domain_size_y,
                domain_size_z,
                mass_maximum,
                time_interval,
                random_seed);

        } /* if ((timestep % timesteps_between_outputs) == 0) */

    } /* for timestep */

   /*
    * If the particles at the last timestep haven't already been
    * output, output them.
    */
    if ((number_of_timesteps % timesteps_between_outputs) != 0) {

        ParticleArray_output(
            base_filename,
            particle_array,
            number_of_particles,
            timestep,
            number_of_timesteps,
            timesteps_between_outputs,
            domain_size_x,
            domain_size_y,
            domain_size_z,
            mass_maximum,
            time_interval,
            random_seed);

    } /* if ((number_of_timesteps % timesteps_between_outputs) != 0) */

   /*
    * Delete the array of forces.
    */
    total_force_array =
        ForceArray_destruct(total_force_array, number_of_particles);

   /*
    * Delete the array of particles.
    */
    particle_array =
        ParticleArray_destruct(particle_array, number_of_particles);

    return program_success_code;

} /* main */

/*
 * Get command line arguments.
 */
void input_arguments (char* arguments_filename, char* base_filename)
{ /* input_arguments */
    const int file_was_closed_code = 0;
    FILE* fileptr = (FILE*)NULL;

    number_of_particles       = default_number_of_particles;
    domain_size_x             = default_domain_size_x;
    domain_size_y             = default_domain_size_y;
    domain_size_z             = default_domain_size_z;
    time_interval             = default_time_interval;
    number_of_timesteps       = default_number_of_timesteps;
    timesteps_between_outputs = default_timesteps_between_outputs;
    random_seed               = default_random_seed;
    mass_maximum              = default_mass_maximum;
    strcpy(base_filename,       default_base_filename);

    if ((arguments_filename == (char*)NULL) ||
        (strlen(arguments_filename) < 1)) {

        fprintf(stderr,
            "ERROR in input_arguments: arguments filename doesn't exist\n");

        die_cleanly();

    } /* if ((arguments_filename == (char*)NULL) || ... ) */

    fileptr = fopen(arguments_filename, "r");

    if (fileptr == (FILE*)NULL) {

        fprintf(stderr,
            "ERROR in input_arguments: can't open arguments file %s\n",
            arguments_filename);

        die_cleanly();

    } /* if (fileptr == (FILE*)NULL) */

    if (fscanf(fileptr, "%d", &number_of_particles) != 1) {

        fprintf(stderr,
            "ERROR: cannot read number of particles from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%d", &number_of_particles) != 1) */

    if (number_of_particles < 1) {

        fprintf(stderr, "ERROR: cannot have %d particles!\n",
            number_of_particles);

        die_cleanly();

    } /* if (number_of_particles < 1) */

    if (number_of_particles == 1) {

        fprintf(stderr,
            "There is only one particle, therefore no forces.\n");

        die_cleanly();

    } /* if (number_of_particles == 1) */

    if (fscanf(fileptr, "%f", &domain_size_x) != 1) {

        fprintf(stderr,
            "ERROR: cannot read domain size X from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%f", &domain_size_x) != 1) */

    if (domain_size_x <= 0.0) {

        fprintf(stderr,
            "ERROR: cannot have a domain whose X dimension has length %f!\n",
            domain_size_x);

        die_cleanly();

    } /* if (domain_size_x <= 0.0) */

    if (fscanf(fileptr, "%f", &domain_size_y) != 1) {

        fprintf(stderr,
            "ERROR: cannot read domain size Y from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%f", &domain_size_y) != 1) */

    if (domain_size_y <= 0.0) {

        fprintf(stderr,
            "ERROR: cannot have a domain whose Y dimension has length %f!\n",
            domain_size_y);

        die_cleanly();

    } /* if (domain_size_y <= 0.0) */

    if (fscanf(fileptr, "%f", &domain_size_z) != 1) {

        fprintf(stderr,
            "ERROR: cannot read domain size Z from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%f", &domain_size_z) != 1) */

    if (domain_size_z <= 0.0) {

        fprintf(stderr,
            "ERROR: cannot have a domain whose Z dimension has length %f!\n",
            domain_size_z);

        die_cleanly();

    } /* if (domain_size_z <= 0.0) */

    if (fscanf(fileptr, "%f", &time_interval) != 1) {

        fprintf(stderr,
            "ERROR: cannot read time interval from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%f", &time_interval) != 1) */

    if (time_interval <= 0.0) {

        fprintf(stderr,
            "ERROR: cannot have a time interval of %f!\n",
            time_interval);

        die_cleanly();

    } /* if (time_interval <= 0.0) */

    if (fscanf(fileptr, "%d", &number_of_timesteps) != 1) {

        fprintf(stderr,
            "ERROR: cannot read number of timesteps from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%d", &number_of_timesteps) != 1) */

    if (number_of_timesteps <= 0) {

        fprintf(stderr, "ERROR: cannot have %d timesteps!\n",
            number_of_timesteps);

        die_cleanly();

    } /* if (number_of_timesteps <= 0) */

    if (fscanf(fileptr, "%d", &timesteps_between_outputs) != 1) {

        fprintf(stderr,
        "ERROR: cannot read timesteps between outputs from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%d", &timesteps_between_outputs) != 1) */

    if (timesteps_between_outputs <= 0) {

        fprintf(stderr,
            "ERROR: cannot have %d timesteps between outputs!\n",
            timesteps_between_outputs);

        die_cleanly();

    } /* if (timesteps_between_outputs <= 0) */

    if (fscanf(fileptr, "%d", &random_seed) != 1) {

        fprintf(stderr,
            "ERROR: cannot read random seed from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%d", &random_seed) != 1) */

    if (fscanf(fileptr, "%f", &mass_maximum) != 1) {

        fprintf(stderr,
            "ERROR: cannot read mass maximum from standard input!\n");

        die_cleanly();

    } /* if (fscanf(fileptr, "%f", &mass_maximum) != 1) */

    if (mass_maximum <= 0.0) {

        fprintf(stderr,
            "ERROR: cannot have a maximum mass of %f!\n",
            mass_maximum);

        die_cleanly();

    } /* if (mass_maximum <= 0.0) */

    getc(fileptr);
    fgets(base_filename, default_string_length, fileptr);
    if (base_filename[strlen(base_filename)-1] == '\n') {
        base_filename[strlen(base_filename)-1] = '\0';
    } /* if (base_filename[strlen(base_filename)-1]...) */

    if (fclose(fileptr) != file_was_closed_code) {

        fprintf(stderr,
          "ERROR in input_arguments: can't close arguments file %s\n",
          arguments_filename);

        die_cleanly();

    } /* if (fclose(fileptr) != file_was_closed_code) */

} /* input_arguments */

