
/* nbody_force.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nbody_force.h"

#ifdef CHECK
static void Force_check(Force this_force, char* action, char* routine);
#endif /* #ifdef CHECK */

static void Force_clear(Force this_force);
static void Force_construct(Force this_force);
static void Force_destruct(Force this_force);

#ifdef CHECK
static void ForceArray_check(
                ForceArray this_force_array, int number_of_forces,
                char* action, char* routine);
#endif /* #ifdef CHECK */

static ForceArray ForceArray_allocate(int number_of_forces);
static ForceArray ForceArray_deallocate(
                      ForceArray this_force_array,
                      int number_of_forces);

#ifdef CHECK
/*
 * Check that the force exists.
 */
void Force_check (Force this_force, char* action, char* routine)
{ /* Force_check */

    if (this_force != (float*)NULL) return;

    fprintf(stderr, "ERROR: can't %s a nonexistent force\n",
        ((action == (char*)NULL) || (strlen(action) == 0)) ?
            "perform an unknown action on" : action);
    fprintf(stderr, "  in %s\n",
        ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
            "an unknown routine" : routine);

    die_cleanly();

} /* Force_check */
#endif /* #ifdef CHECK */

/*
 * Clear the force's data.
 */
void Force_clear (Force this_force)
{ /* Force_clear */

#ifdef CHECK
    Force_check(this_force, "clear", "Force_clear");
#endif /* #ifdef CHECK */

    this_force[force_x_index] = 0.0;
    this_force[force_y_index] = 0.0;
    this_force[force_z_index] = 0.0;

} /* Force_clear */

/*
 * Construct the force.
 */
void Force_construct (Force this_force)
{ /* Force_construct */

#ifdef CHECK
    Force_check(this_force, "construct", "Force_construct");
#endif /* #ifdef CHECK */

    Force_clear(this_force);

} /* Force_construct */

/*
 * Destroy the force.
 */
void Force_destruct (Force this_force)
{ /* Force_destruct */

#ifdef CHECK
    Force_check(this_force, "destruct", "Force_destruct");
#endif /* #ifdef CHECK */

    Force_clear(this_force);

} /* Force_destruct */

#ifdef CHECK
void ForceArray_check (
         ForceArray this_force_array, int number_of_forces,
         char* action, char* routine)
{ /* ForceArray_check */

    if (number_of_forces < 0) {

        fprintf(stderr, "ERROR: illegal number of forces %d\n",
            number_of_forces);
        fprintf(stderr, "  to %s\n", 
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (number_of_forces < 0) */

    if (number_of_forces == 0) {

        if (this_force_array == (ForceArray)NULL) {
            return (ForceArray)NULL;
        } /* if (this_force_array == (ForceArray)NULL) */

        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  an existing force array of length 0\n");
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (number_of_forces == 0) */
    if (this_force_array == (ForceArray)NULL) {

        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  a nonexistent array of %d forces\n",
            number_of_forces);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (this_force_array == (ForceArray)NULL) */

} /* ForceArray_check */
#endif /* #ifdef CHECK */

/*
 * Allocate and return an array of forces.
 */
ForceArray ForceArray_allocate (int number_of_forces)
{ /* ForceArray_allocate */
    ForceArray this_force_array = (ForceArray)NULL;
    int index;

    if (number_of_forces < 0) {

        fprintf(stderr,
            "ERROR: illegal number of forces %d to allocate\n", 
            number_of_forces);
        fprintf(stderr, "  in ForceArray_construct\n");

        die_cleanly();

    } /* if (number_of_forces < 0) */

    if (number_of_forces == 0) return (ForceArray)NULL;

    this_force_array =
        (ForceArray)malloc(sizeof(Force) * number_of_forces);

    if (this_force_array == (ForceArray)NULL) {

        fprintf(stderr,
            "ERROR: can't allocate a force array of %d forces\n", 
            number_of_forces);
        fprintf(stderr, "  in ForceArray_construct\n");

        die_cleanly();

    } /* if (this_force_array == (ForceArray)NULL) */

    return this_force_array;

} /* ForceArray_allocate */

/*
 * Construct and return an array of forces, cleared.
 */
ForceArray ForceArray_construct (int number_of_forces)
{ /* ForceArray_construct */
    ForceArray this_force_array = (ForceArray)NULL;
    int index;

    this_force_array = ForceArray_allocate(number_of_forces);

    for (index = 0; index < number_of_forces; index++) {

        Force_construct(this_force_array[index]);

    } /* for index */

    return this_force_array;

} /* ForceArray_construct */

/*
 * Deallocate the array of forces, and return NULL.
 */
ForceArray ForceArray_deallocate (
                  ForceArray this_force_array,
                  int number_of_forces)
{ /* ForceArray_deallocate */
    int index;

#ifdef CHECK
    ForceArray_check(this_force_array, number_of_forces,
        "deallocate", "ForceArray_deallocate");
#endif /* #ifdef CHECK */

    free(this_force_array);

    this_force_array = (ForceArray)NULL;

    return (ForceArray)NULL;

} /* ForceArray_deallocate */

/*
 * Destroy the array of forces, and return NULL.
 */
ForceArray ForceArray_destruct (
                  ForceArray this_force_array,
                  int number_of_forces)
{ /* ForceArray_destruct */
    int index;

#ifdef CHECK
    ForceArray_check(this_force_array, number_of_forces,
        "destroy", "ForceArray_destruct");
#endif /* #ifdef CHECK */

    for (index = number_of_forces - 1; index >= 0; index--) {

        Force_destruct(this_force_array[index]);

    } /* for index */

    return ForceArray_deallocate(this_force_array, number_of_forces);

} /* ForceArray_destruct */

/*
 * Calculate the force between two particles.
 *
 * Gregory R. Andrews, "Foundations of Multithreaded, Parallel, and
 * Distributed Programming," Addison-Wesley, 2000, p. 554-6.
 */
void Force_calculate (Force this_force,
                      Particle this_particle1, Particle this_particle2)
{ /* Force_calculate */
    float difference_x, difference_y, difference_z;
    float distance_squared, distance;
    float force_magnitude;

#ifdef CHECK
    Particle_pair_check(this_particle1, this_particle2,
        "calculate the force along the X-axis", "Force_calculate_x");
#endif /* #ifdef CHECK */

    difference_x =
        this_particle2[position_x_index] - this_particle1[position_x_index];
    difference_y =
        this_particle2[position_y_index] - this_particle1[position_y_index];
    difference_z =
        this_particle2[position_z_index] - this_particle1[position_z_index];

    distance_squared = difference_x * difference_x +
                       difference_y * difference_y +
                       difference_z * difference_z;

    distance = sqrt(distance_squared);

    force_magnitude =
        gravitational_constant *
        (this_particle1[mass_index]) * (this_particle2[mass_index]) /
        distance_squared;

    this_force[force_x_index] =
        (force_magnitude / distance) * difference_x;
    this_force[force_y_index] =
        (force_magnitude / distance) * difference_y;
    this_force[force_z_index] =
        (force_magnitude / distance) * difference_z;

} /* Force_calculate */

/*
 * Calculate the total force on each particle by calculating
 * the pairwise forces between every pair of particles.
 */
void ForceArray_calculate (
         ForceArray this_total_force_array,
         ParticleArray this_particle_array, int number_of_particles)
{ /* ForceArray_calculate */
    Force pair_force;
    int   index1, index2;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "calculate forces", "ForceArray_calculate");
#endif /* #ifdef CHECK */

   /*
    * In a 1-body problem, there aren't any forces to calculate.
    */
    if (number_of_particles <= 1) return;

    for (index1 = 0; index1 < number_of_particles; index1++) {

       /*
        * Initialize the total force on particle index1.
        */
        this_total_force_array[index1][force_x_index] = 0.0;
        this_total_force_array[index1][force_y_index] = 0.0;
        this_total_force_array[index1][force_z_index] = 0.0;

        for (index2 = 0; index2 < number_of_particles; index2++) {

           /*
            * The force between a particle and itself is zero,
            * so there's no need to calculate it.
            */
            if (index1 != index2) {

               /*
                * Calculate the force between
                * particle index1 and particle index2.
                */
                Force_calculate(
                    pair_force,
                    this_particle_array[index1],
                    this_particle_array[index2]);

               /*
                * Increase the total force on particle index1 by
                * the force between particle index1 and particle index2.
                */
                this_total_force_array[index1][force_x_index] +=
                    pair_force[force_x_index];
                this_total_force_array[index1][force_y_index] +=
                    pair_force[force_y_index];
                this_total_force_array[index1][force_z_index] +=
                    pair_force[force_z_index];

            } /* if (index1 != index2) */

        } /* for index2 */

    } /* for index1 */

} /* ForceArray_calculate */

