
/* nbody_particle.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nbody_particle.h"

#ifdef CHECK
static void Particle_check(
                Particle this_particle, char* action, char* routine);
static void Particle_pair_check(
                Particle this_particle1, Particle this_particle2,
                char* action, char* routine);
#endif /* #ifdef CHECK */

static void Particle_clear(Particle this_particle);
static void Particle_construct(Particle this_particle);
static void Particle_destruct(Particle this_particle);
static void Particle_set_position_randomly(
                Particle this_particle,
                float domain_size_x,
                float domain_size_y,
                float domain_size_z);
static void Particle_initialize_randomly(
                Particle this_particle,
                float domain_size_x,
                float domain_size_y,
                float domain_size_z,
                float mass_maximum);
static void Particle_initialize(
                Particle this_particle,
                float domain_size_x,
                float domain_size_y,
                float domain_size_z,
                float mass_maximum);
static void Particle_move(Particle this_particle, Force this_force,
                          float time_interval);
static void Particle_output(FILE* fileptr, Particle this_particle);

#ifdef CHECK
static void ParticleArray_check(
                ParticleArray this_particle_array, int number_of_particles,
                char* action, char* routine);
#endif /* #ifdef CHECK */

static ParticleArray ParticleArray_allocate(int number_of_particles);
static ParticleArray ParticleArray_deallocate(
                         ParticleArray this_particle_array,
                         int number_of_particles);
static void ParticleArray_initialize_randomly(
                ParticleArray this_particle_array,
                int number_of_particles,
                float domain_size_x,
                float domain_size_y,
                float domain_size_z,
                float mass_maximum);

#ifdef CHECK
/*
 * Check that the particle exists.
 */
void Particle_check (Particle this_particle, char* action, char* routine)
{ /* Particle_check */

    if (this_particle != (float*)NULL) return;

    fprintf(stderr, "ERROR: can't %s a nonexistent particle\n",
        ((action == (char*)NULL) || (strlen(action) == 0)) ?
            "perform an unknown action on" : action);
    fprintf(stderr, "  in %s\n",
        ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
            "an unknown routine" : routine);

    die_cleanly();

} /* Particle_check */

/*
 * Check that the pair of particles both exist.
 */
void Particle_pair_check (Particle this_particle1, Particle this_particle2,
         char* action, char* routine)
{ /* Particle_pair_check */

    if ((this_particle1 != (float*)NULL) &&
        (this_particle2 != (float*)NULL))
        return;

    fprintf(stderr, "ERROR: can't %s\n",
        ((action == (char*)NULL) || (strlen(action) == 0)) ?
            "perform an unknown action on" : action);

    if (this_particle1 == (float*)NULL) {

        if (this_particle2 == (float*)NULL) {

            fprintf(stderr, "  a pair of nonexistent particles\n");

        } /* if (this_particle2 == (float*)NULL) */
        else {

            fprintf(stderr,
                "  a nonexistent particle and an existing particle\n");

        } /* if (this_particle2 == (float*)NULL)...else */

    } /* if (this_particle1 == (float*)NULL) */
    else {

        fprintf(stderr,
            "  an existing particle and a nonexistent particle\n");

    } /* if (this_particle1 == (float*)NULL)...else */

    fprintf(stderr, "  in %s\n",
        ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
            "an unknown routine" : routine);

    die_cleanly();

} /* Particle_pair_check */
#endif /* #ifdef CHECK */

/*
 * Clear the particle's data.
 */
void Particle_clear (Particle this_particle)
{ /* Particle_clear */

#ifdef CHECK
    Particle_check(this_particle, "clear", "Particle_clear");
#endif /* #ifdef CHECK */

    this_particle[position_x_index] = 0.0;
    this_particle[position_y_index] = 0.0;
    this_particle[position_z_index] = 0.0;
    this_particle[velocity_x_index] = 0.0;
    this_particle[velocity_y_index] = 0.0;
    this_particle[velocity_z_index] = 0.0;
    this_particle[mass_index]       = 0.0;

} /* Particle_clear */

/*
 * Construct the particle.
 */
void Particle_construct (Particle this_particle)
{ /* Particle_construct */

#ifdef CHECK
    Particle_check(this_particle, "construct", "Particle_construct");
#endif /* #ifdef CHECK */

    Particle_clear(this_particle);

} /* Particle_construct */

/*
 * Destroy the particle.
 */
void Particle_destruct (Particle this_particle)
{ /* Particle_destruct */

#ifdef CHECK
    Particle_check(this_particle, "destruct", "Particle_destruct");
#endif /* #ifdef CHECK */

    Particle_clear(this_particle);

} /* Particle_destruct */

/*
 * Initialize the particle by setting its data randomly.
 */
void Particle_set_position_randomly (Particle this_particle,
                                     float domain_size_x,
                                     float domain_size_y,
                                     float domain_size_z)
{ /* Particle_set_position_randomly */

#ifdef CHECK
    Particle_check(this_particle, "randomly set the position",
        "Particle_set_randomly");
#endif /* #ifdef CHECK */

    this_particle[position_x_index] = domain_size_x * uniform_random_0_to_1();
    this_particle[position_y_index] = domain_size_y * uniform_random_0_to_1();
    this_particle[position_z_index] = domain_size_z * uniform_random_0_to_1();

} /* Particle_set_position_randomly */

/*
 * Initialize the particle by setting its data randomly.
 */
void Particle_initialize_randomly (Particle this_particle,
                                   float domain_size_x,
                                   float domain_size_y,
                                   float domain_size_z,
                                   float mass_maximum)
{ /* Particle_initialize_randomly */

#ifdef CHECK
    Particle_check(this_particle, "randomly initialize",
        "Particle_initialize_randomly");
#endif /* #ifdef CHECK */

    Particle_clear(this_particle);

    Particle_set_position_randomly(
        this_particle,
        domain_size_x, domain_size_y, domain_size_z);

    this_particle[mass_index] = mass_maximum * uniform_random_0_to_1();

} /* Particle_initialize_randomly */

/*
 * Initialize the particle.
 */
void Particle_initialize (Particle this_particle,
                          float domain_size_x,
                          float domain_size_y,
                          float domain_size_z,
                          float mass_maximum)
{ /* Particle_initialize */

#ifdef CHECK
    Particle_check(this_particle, "initialize", "Particle_initialize");
#endif /* #ifdef CHECK */

    Particle_initialize_randomly(
        this_particle,
        domain_size_x, domain_size_y, domain_size_z, mass_maximum);

} /* Particle_initialize */

/*
 * Move the particle based on the force applied to it.
 *
 * Gregory R. Andrews, "Foundations of Multithreaded, Parallel, and
 * Distributed Programming," Addison-Wesley, 2000, p. 554-6.
 */
void Particle_move (Particle this_particle, Force this_force,
                    float time_interval)
{ /* Particle_move */
    float this_mass;
    float velocity_change_x, velocity_change_y, velocity_change_z;
    float position_change_x, position_change_y, position_change_z;

#ifdef CHECK
    Particle_check(this_particle, "move", "Particle_move");
#endif /* #ifdef CHECK */

    this_mass = this_particle[mass_index];

   /*
    * Calculate the change in each component of velocity.
    */

    velocity_change_x =
        this_force[force_x_index] * (time_interval / this_mass);
    velocity_change_y =
        this_force[force_y_index] * (time_interval / this_mass);
    velocity_change_z =
        this_force[force_z_index] * (time_interval / this_mass);

   /*
    * Calculate the change in each component of position.
    */

    position_change_x =
        this_particle[velocity_x_index] +
        velocity_change_x * (0.5 * time_interval);
    position_change_y =
        this_particle[velocity_y_index] +
        velocity_change_y * (0.5 * time_interval);
    position_change_z =
        this_particle[velocity_z_index] +
        velocity_change_z * (0.5 * time_interval);

   /*
    * Calculate the new value of each component of velocity.
    */

    this_particle[velocity_x_index] += velocity_change_x;
    this_particle[velocity_y_index] += velocity_change_y;
    this_particle[velocity_z_index] += velocity_change_z;

   /*
    * Calculate the new value of each component of position.
    */

    this_particle[position_x_index] += position_change_x;
    this_particle[position_y_index] += position_change_y;
    this_particle[position_z_index] += position_change_z;

} /* Particle_move */

void Particle_output (FILE* fileptr, Particle this_particle)
{ /* Particle_output */

    fprintf(fileptr, "%g %g %g %g %g %g %g\n",
        this_particle[position_x_index],
        this_particle[position_y_index],
        this_particle[position_z_index],
        this_particle[velocity_x_index],
        this_particle[velocity_y_index],
        this_particle[velocity_z_index],
        this_particle[mass_index]);

} /* Particle_output */

#ifdef CHECK
void ParticleArray_check (
         ParticleArray this_particle_array, int number_of_particles,
         char* action, char* routine)
{ /* ParticleArray_check */

    if (number_of_particles < 0) {

        fprintf(stderr, "ERROR: illegal number of particles %d\n",
            number_of_particles);
        fprintf(stderr, "  to %s\n", 
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (number_of_particles < 0) */

    if (number_of_particles == 0) {

        if (this_particle_array == (ParticleArray)NULL) {
            return (ParticleArray)NULL;
        } /* if (this_particle_array == (ParticleArray)NULL) */

        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  an existing particle array of length 0\n");
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (number_of_particles == 0) */
    if (this_particle_array == (ParticleArray)NULL) {

        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  a nonexistent array of %d particles\n",
            number_of_particles);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);

        die_cleanly();

    } /* if (this_particle_array == (ParticleArray)NULL) */

} /* ParticleArray_check */
#endif /* #ifdef CHECK */

/*
 * Allocate and return an array of particles.
 */
ParticleArray ParticleArray_allocate (int number_of_particles)
{ /* ParticleArray_allocate */
    ParticleArray this_particle_array = (ParticleArray)NULL;
    int index;

    if (number_of_particles < 0) {

        fprintf(stderr,
            "ERROR: illegal number of particles %d to allocate\n", 
            number_of_particles);
        fprintf(stderr, "  in ParticleArray_construct\n");

        die_cleanly();

    } /* if (number_of_particles < 0) */

    if (number_of_particles == 0) return (ParticleArray)NULL;

    this_particle_array =
        (ParticleArray)malloc(sizeof(Particle) * number_of_particles);

    if (this_particle_array == (ParticleArray)NULL) {

        fprintf(stderr,
            "ERROR: can't allocate a particle array of %d particles\n", 
            number_of_particles);
        fprintf(stderr, "  in ParticleArray_construct\n");

        die_cleanly();

    } /* if (this_particle_array == (ParticleArray)NULL) */

    return this_particle_array;

} /* ParticleArray_allocate */

/*
 * Construct and return an array of particles, cleared.
 */
ParticleArray ParticleArray_construct (int number_of_particles)
{ /* ParticleArray_construct */
    ParticleArray this_particle_array = (ParticleArray)NULL;
    int index;

    this_particle_array = ParticleArray_allocate(number_of_particles);

    for (index = 0; index < number_of_particles; index++) {

        Particle_construct(this_particle_array[index]);

    } /* for index */

    return this_particle_array;

} /* ParticleArray_construct */

/*
 * Deallocate the array of particles, and return NULL.
 */
ParticleArray ParticleArray_deallocate (
                  ParticleArray this_particle_array,
                  int number_of_particles)
{ /* ParticleArray_deallocate */
    int index;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "deallocate", "ParticleArray_deallocate");
#endif /* #ifdef CHECK */

    free(this_particle_array);

    this_particle_array = (ParticleArray)NULL;

    return (ParticleArray)NULL;

} /* ParticleArray_deallocate */

/*
 * Destroy the array of particles, and return NULL.
 */
ParticleArray ParticleArray_destruct (
                  ParticleArray this_particle_array,
                  int number_of_particles)
{ /* ParticleArray_destruct */
    int index;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "destroy", "ParticleArray_destruct");
#endif /* #ifdef CHECK */

    for (index = number_of_particles - 1; index >= 0; index--) {

        Particle_destruct(this_particle_array[index]);

    } /* for index */

    return ParticleArray_deallocate(this_particle_array, number_of_particles);

} /* ParticleArray_destruct */

/*
 * Initialize the array of particles by setting its data randomly.
 */
void ParticleArray_initialize_randomly (
         ParticleArray this_particle_array,
         int number_of_particles,
         float domain_size_x,
         float domain_size_y,
         float domain_size_z,
         float mass_maximum)
{ /* ParticleArray_initialize_randomly */
    int index;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "initialize randomly", "ParticleArray_initialize_randomly");
#endif /* #ifdef CHECK */

    for (index = 0; index < number_of_particles; index++) {

        Particle_initialize_randomly(this_particle_array[index],
            domain_size_x, domain_size_y, domain_size_z, mass_maximum);

    } /* for index */

} /* ParticleArray_initialize_randomly */

/*
 * Initialize the array of particles.
 */
void ParticleArray_initialize (
         ParticleArray this_particle_array,
         int number_of_particles,
         float domain_size_x,
         float domain_size_y,
         float domain_size_z,
         float mass_maximum)
{ /* ParticleArray_initialize */

    ParticleArray_initialize_randomly(
        this_particle_array, number_of_particles,
        domain_size_x, domain_size_y, domain_size_z, mass_maximum);

} /* ParticleArray_initialize */

/*
 * Move every particle in the array of particles.
 */
void ParticleArray_move (
         ParticleArray this_particle_array,
         ForceArray this_total_force_array,
         int number_of_particles, float time_interval)
{ /* ParticleArray_move */
    int index;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "calculate forces", "ForceArray_calculate");
    ForceArray_check(this_total_force_array, number_of_particles,
        "calculate forces", "ForceArray_calculate");
#endif /* #ifdef CHECK */

    if (number_of_particles < 1) return;

    for (index = 0; index < number_of_particles; index++) {

        Particle_move(this_particle_array[index],
                      this_total_force_array[index], time_interval);

    } /* for index */

} /* ParticleArray_move */

void ParticleArray_output (char* base_filename,
         ParticleArray this_particle_array, int number_of_particles,
         int timestep, int number_of_timesteps, int timesteps_between_outputs,
         float domain_size_x, float domain_size_y, float domain_size_z,
         float mass_maximum, float time_interval,
         unsigned int random_seed)
{ /* ParticleArray_output */
    FILE* fileptr  = (FILE*)NULL;
    char* filename = (char*)NULL;
    int   filename_length;
    int   index;

#ifdef CHECK
    ParticleArray_check(this_particle_array, number_of_particles,
        "output", "ParticleArray_output");
#endif /* #ifdef CHECK */

    filename_length = strlen(base_filename) + 1 + 8 + 1 + 3;

    filename = (char*)malloc(sizeof(char) * (filename_length + 1));

    if (filename == (char*)NULL) {

        fprintf(stderr, "ERROR: can't allocate the filename string\n");
        fprintf(stderr, "  %s_%8.8d.txt\n", base_filename, timestep);
        fprintf(stderr, "  in ParticleArray_output\n");

        die_cleanly();

    } /* if (filename == (char*)NULL) */

    sprintf(filename, "%s_%8.8d.txt", base_filename, timestep);

    fileptr = fopen(filename, "w");

    if (fileptr == (FILE*)NULL) {

        fprintf(stderr, "ERROR: can't open the output file named\n");
        fprintf(stderr, "  %s\n", filename);
        fprintf(stderr, "  in ParticleArray_output\n");

        die_cleanly();

    } /* if (fileptr == (FILE*)NULL) */

    fprintf(fileptr, "%d %d %d %g %g %g %g %g %d\n",
        number_of_particles, number_of_timesteps, timesteps_between_outputs,
        domain_size_x, domain_size_y, domain_size_z,
        mass_maximum, time_interval,
        random_seed);

    fprintf(fileptr, "%d\n", timestep);

    for (index = 0; index < number_of_particles; index++) {

        Particle_output(fileptr, this_particle_array[index]);

    } /* for index */

    if (fclose(fileptr) != 0) {

        fprintf(stderr, "ERROR: can't close the output file named\n");
        fprintf(stderr, "  %s\n", filename);
        fprintf(stderr, "  in ParticleArray_output\n");

        die_cleanly();

    } /* if (fclose(fileptr) != 0) */

    fileptr = (FILE*)NULL;

    free(filename); filename = (char*)NULL;

} /* ParticleArray_output */

