!
! nbody_force.f90
!

      MODULE nbody_force

          USE nbody_particle_data

          IMPLICIT NONE

          PUBLIC
          SAVE

          REAL,   PARAMETER :: gravitational_constant = 6.6726e-11 !! N(m/kg)^2

          INTEGER,PARAMETER :: first_force_component_index = 1
          INTEGER,PARAMETER :: force_x_index               =      &
     &                             first_force_component_index
          INTEGER,PARAMETER :: force_y_index               =      &
     &                             (force_x_index + 1)
          INTEGER,PARAMETER :: force_z_index               =      &
     &                             (force_y_index + 1)
          INTEGER,PARAMETER :: last_force_component_index  =      &
     &                              force_z_index
          INTEGER,PARAMETER :: number_of_force_components  =      &
     &                             (last_force_component_index -  &
     &                              first_force_component_index + 1)

      CONTAINS !! MODULE nbody_force

!!!!!!!!!
!!!!!!!!! Calculate the force between two particles.
!!!!!!!!!
!!!!!!!!! Gregory R. Andrews, "Foundations of Multithreaded, Parallel
!!!!!!!!! and Distributed Programming," Addison-Wesley, 2000, p. 554-6.
!!!!!!!!!
          SUBROUTINE Force_calculate (  &
     &                   this_force, this_particle1, this_particle2)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_force_components),     &
     &             INTENT(OUT) :: this_force
              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(IN)  :: this_particle1, this_particle2

              REAL :: difference_x, difference_y, difference_z
              REAL :: distance_squared, distance
              REAL :: force_magnitude

              difference_x =                          &
     &            this_particle2(position_x_index) -  &
     &            this_particle1(position_x_index)
              difference_y =                          &
     &            this_particle2(position_y_index) -  &
     &            this_particle1(position_y_index)
              difference_z =                          &
     &            this_particle2(position_z_index) -  &
     &            this_particle1(position_z_index)

              distance_squared = difference_x * difference_x +  &
     &                           difference_y * difference_y +  &
     &                           difference_z * difference_z

              distance = SQRT(distance_squared)

              force_magnitude =                   &
     &            gravitational_constant *        &
     &            (this_particle1(mass_index) *   &
     &             this_particle2(mass_index)) /  &
     &            distance_squared

              this_force(force_x_index) =  &
     &            (force_magnitude / distance) * difference_x
              this_force(force_y_index) =  &
     &            (force_magnitude / distance) * difference_y
              this_force(force_z_index) =  &
     &            (force_magnitude / distance) * difference_z

          END SUBROUTINE Force_calculate

!!!!!!!!!
!!!!!!!!! Calculate the total force on each particle by calculating
!!!!!!!!! the pairwise forces between every pair of particles.
!!!!!!!!!
          SUBROUTINE ForceArray_calculate (                           &
     &                   this_total_force_array, this_particle_array, &
     &                   number_of_particles)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_force_components,          &
     &                       number_of_particles),INTENT(OUT) ::  &
     &            this_total_force_array
              REAL,DIMENSION(number_of_particle_components,       &
     &                       number_of_particles),INTENT(IN) ::   &
     &            this_particle_array
              INTEGER,INTENT(IN) :: number_of_particles

              REAL,DIMENSION(number_of_force_components) :: pair_force
              INTEGER :: index1, index2

!............
!............ In a 1-body problem, there aren't any forces to calculate.
!............
              IF (number_of_particles <= 1) THEN

                  RETURN

              END IF !! (number_of_particles <= 1)

              DO index1 = 1, number_of_particles

!................
!................ Initialize the total force on particle index1.
!................
                  this_total_force_array(force_x_index,index1) = 0.0
                  this_total_force_array(force_y_index,index1) = 0.0
                  this_total_force_array(force_z_index,index1) = 0.0

                  DO index2 = 1, number_of_particles

!....................
!.................... The force between a particle and itself is zero,
!.................... so there's no need to calculate it.
!....................
                      IF (index1 /= index2) THEN

!........................
!........................ Calculate the force between
!........................ particle index1 and particle index2.
!........................
                          CALL Force_calculate(                    &
     &                             pair_force,                     &
     &                             this_particle_array(:,index1),  &
     &                             this_particle_array(:,index2))

!........................
!........................ Increase the total force on particle index1
!........................ by the force between particle index1 and
!........................ particle index2.
!........................
                          this_total_force_array(force_x_index,      &
     &                                           index1) =           &
     &                        this_total_force_array(force_x_index,  &
     &                                               index1) +       &
     &                        pair_force(force_x_index)
                          this_total_force_array(force_y_index,      &
     &                                           index1) =           &
     &                        this_total_force_array(force_y_index,  &
     &                                               index1) +       &
     &                        pair_force(force_y_index)
                          this_total_force_array(force_z_index,      &
     &                                           index1) =           &
     &                        this_total_force_array(force_z_index,  &
     &                                               index1) +       &
     &                        pair_force(force_z_index)

                      END IF !! (index1 /= index2)

                  END DO !! index2

              END DO !! index1

          END SUBROUTINE ForceArray_calculate

      END MODULE nbody_force

