!
! nbody_particle.f90
!

      MODULE nbody_particle

          USE nbody_force

          IMPLICIT NONE

          PUBLIC
          SAVE

      CONTAINS !! MODULE nbody_particle

!!!!!!!!!
!!!!!!!!! Initialize the particle by setting its data randomly.
!!!!!!!!!
          SUBROUTINE Particle_set_position_randomly (  &
     &                   this_particle,                &
     &                   domain_size_x, domain_size_y, domain_size_z)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(OUT) :: this_particle
              REAL,INTENT(IN)  :: domain_size_x
              REAL,INTENT(IN)  :: domain_size_y
              REAL,INTENT(IN)  :: domain_size_z

              this_particle(position_x_index) =  &
     &            domain_size_x * uniform_random_0_to_1()
              this_particle(position_y_index) =  &
     &            domain_size_y * uniform_random_0_to_1()
              this_particle(position_z_index) =  &
     &            domain_size_z * uniform_random_0_to_1()

          END SUBROUTINE Particle_set_position_randomly

!!!!!!!!!
!!!!!!!!! Initialize the particle by setting its data randomly.
!!!!!!!!!
          SUBROUTINE Particle_initialize_randomly (  &
     &                   this_particle,              &
     &                   domain_size_x,              &
     &                   domain_size_y,              &
     &                   domain_size_z,              &
     &                   mass_maximum)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(OUT) :: this_particle
              REAL,INTENT(IN)  :: domain_size_x
              REAL,INTENT(IN)  :: domain_size_y
              REAL,INTENT(IN)  :: domain_size_z
              REAL,INTENT(IN)  :: mass_maximum

              CALL Particle_set_position_randomly(  &
     &                 this_particle,               &
     &                 domain_size_x, domain_size_y, domain_size_z)

              this_particle(mass_index) =  &
     &            mass_maximum * uniform_random_0_to_1()

          END SUBROUTINE Particle_initialize_randomly

!!!!!!!!!
!!!!!!!!! Initialize the particle.
!!!!!!!!!
          SUBROUTINE Particle_initialize (  &
     &                   this_particle,     &
     &                   domain_size_x,     &
     &                   domain_size_y,     &
     &                   domain_size_z,     &
     &                   mass_maximum)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(OUT) :: this_particle
              REAL,INTENT(IN)  :: domain_size_x
              REAL,INTENT(IN)  :: domain_size_y
              REAL,INTENT(IN)  :: domain_size_z
              REAL,INTENT(IN)  :: mass_maximum

              CALL Particle_initialize_randomly(                     &
     &                 this_particle,                                &
     &                 domain_size_x, domain_size_y, domain_size_z,  &
     &                 mass_maximum)

          END SUBROUTINE Particle_initialize

!!!!!!!!!
!!!!!!!!! Move the particle based on the force applied to it.
!!!!!!!!!
!!!!!!!!! Gregory R. Andrews, "Foundations of Multithreaded, Parallel,
!!!!!!!!! and Distributed Programming," Addison-Wesley, 2000, p. 554-6.
!!!!!!!!!
          SUBROUTINE Particle_move (this_particle, this_force,  &
     &                              time_interval)
              IMPLICIT NONE

              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(INOUT) :: this_particle
              REAL,DIMENSION(number_of_force_components),     &
     &             INTENT(IN)    :: this_force
              REAL,INTENT(IN)    :: time_interval

              REAL :: this_mass
              REAL :: velocity_change_x
              REAL :: velocity_change_y
              REAL :: velocity_change_z
              REAL :: position_change_x
              REAL :: position_change_y
              REAL :: position_change_z

              this_mass = this_particle(mass_index)

!............
!............ Calculate the change in each component of velocity.
!............

              velocity_change_x =              &
     &            this_force(force_x_index) *  &
     &            (time_interval / this_mass)
              velocity_change_y =              &
     &            this_force(force_y_index) *  &
     &            (time_interval / this_mass)
              velocity_change_z =              &
     &            this_force(force_z_index) *  &
     &            (time_interval / this_mass)

!............
!............ Calculate the change in each component of position.
!............

              position_change_x =                    &
     &            this_particle(velocity_x_index) +  &
     &            velocity_change_x * (0.5 * time_interval)
              position_change_y =                    &
     &            this_particle(velocity_y_index) +  &
     &            velocity_change_y * (0.5 * time_interval)
              position_change_z =                    &
     &            this_particle(velocity_z_index) +  &
     &            velocity_change_z * (0.5 * time_interval)

!............
!............ Calculate the new value of each component of velocity.
!............

              this_particle(velocity_x_index) =      &
     &            this_particle(velocity_x_index) +  &
     &            velocity_change_x
              this_particle(velocity_y_index) =      &
     &            this_particle(velocity_y_index) +  &
     &            velocity_change_y
              this_particle(velocity_z_index) =      &
     &            this_particle(velocity_z_index) +  &
     &            velocity_change_z

!............
!............ Calculate the new value of each component of position.
!............

              this_particle(position_x_index) =      &
     &        this_particle(position_x_index) +  &
     &        position_change_x
              this_particle(position_y_index) =      &
     &        this_particle(position_y_index) +  &
     &        position_change_y
              this_particle(position_z_index) =      &
                  this_particle(position_z_index) +  &
     &        position_change_z

          END SUBROUTINE Particle_move

          SUBROUTINE Particle_output (output_unit, this_particle)
              IMPLICIT NONE

              INTEGER,INTENT(IN) :: output_unit
              REAL,DIMENSION(number_of_particle_components),  &
     &             INTENT(IN) :: this_particle

              WRITE (output_unit,                                 &
     &            "(E11.3,E11.3,E11.3,E11.3,E11.3,E11.3,E11.3)")  &
     &            this_particle(position_x_index),                &
     &            this_particle(position_y_index),                &
     &            this_particle(position_z_index),                &
     &            this_particle(velocity_x_index),                &
     &            this_particle(velocity_y_index),                &
     &            this_particle(velocity_z_index),                &
     &            this_particle(mass_index)

          END SUBROUTINE Particle_output

!!!!!!!!!
!!!!!!!!! Initialize the array of particles by setting its data randomly.
!!!!!!!!!
          SUBROUTINE ParticleArray_initialize_randomly (  &
     &                   this_particle_array,             &
     &                   number_of_particles,             &
     &                   domain_size_x,                   &
     &                   domain_size_y,                   &
     &                   domain_size_z,                   &
     &                   mass_maximum)
              IMPLICIT NONE

              INTEGER,INTENT(IN) :: number_of_particles
              REAL,DIMENSION(number_of_particle_components,  &
     &                       number_of_particles),           &
     &             INTENT(OUT) :: this_particle_array
              REAL,   INTENT(IN) :: domain_size_x
              REAL,   INTENT(IN) :: domain_size_y
              REAL,   INTENT(IN) :: domain_size_z
              REAL,   INTENT(IN) :: mass_maximum

              INTEGER :: index

              DO index = 1, number_of_particles

                  CALL Particle_initialize_randomly(                 &
     &                 this_particle_array(:,index),                 &
     &                 domain_size_x, domain_size_y, domain_size_z,  &
     &                 mass_maximum)

              END DO !! index

          END SUBROUTINE ParticleArray_initialize_randomly

!!!!!!!!!
!!!!!!!!! Initialize the array of particles.
!!!!!!!!!
          SUBROUTINE ParticleArray_initialize (  &
     &                   this_particle_array,    &
     &                   number_of_particles,    &
     &                   domain_size_x,          &
     &                   domain_size_y,          &
     &                   domain_size_z,          &
     &                   mass_maximum)
              IMPLICIT NONE

              INTEGER,INTENT(IN) :: number_of_particles
              REAL,DIMENSION(number_of_particle_components,  &
     &                       number_of_particles),           &
     &             INTENT(OUT) :: this_particle_array
              REAL,INTENT(IN) :: domain_size_x
              REAL,INTENT(IN) :: domain_size_y
              REAL,INTENT(IN) :: domain_size_z
              REAL,INTENT(IN) :: mass_maximum

              CALL ParticleArray_initialize_randomly(                &
     &                 this_particle_array, number_of_particles,     &
     &                 domain_size_x, domain_size_y, domain_size_z,  &
     &                 mass_maximum)

          END SUBROUTINE ParticleArray_initialize

!!!!!!!!!
!!!!!!!!! Move every particle in the array of particles.
!!!!!!!!!
          SUBROUTINE ParticleArray_move (         &
     &                   this_particle_array,     &
     &                   this_total_force_array,  &
     &                   number_of_particles, time_interval)
              IMPLICIT NONE

              INTEGER :: number_of_particles
              REAL,DIMENSION(number_of_particle_components,  &
     &                       number_of_particles),           &
     &             INTENT(INOUT) :: this_particle_array
              REAL,DIMENSION(number_of_force_components,     &
     &                       number_of_particles),           &
     &             INTENT(IN)    :: this_total_force_array
              REAL :: time_interval

              INTEGER :: index

              IF (number_of_particles < 1) THEN

                  RETURN

             END IF !! (number_of_particles < 1)

              DO index = 1, number_of_particles

                  CALL Particle_move(                        &
     &                     this_particle_array(:,index),     &
     &                     this_total_force_array(:,index),  &
     &                     time_interval)

              END DO !! index

          END SUBROUTINE ParticleArray_move

          SUBROUTINE ParticleArray_output (          &
     &                   base_filename,              &
     &                   this_particle_array,        &
     &                   number_of_particles,        &
     &                   timestep,                   &
     &                   number_of_timesteps,        &
     &                   timesteps_between_outputs,  &
     &                   domain_size_x,              &
     &                   domain_size_y,              &
     &                   domain_size_z,              &
     &                   mass_maximum,               &
     &                   time_interval,              &
     &                   random_seed_value)
              IMPLICIT NONE

              CHARACTER (LEN=*),INTENT(IN) :: base_filename
              INTEGER,INTENT(IN)           :: number_of_particles
              REAL,DIMENSION(number_of_particle_components,  &
     &                       number_of_particles),           &
     &             INTENT(IN) :: this_particle_array
              INTEGER,INTENT(IN)           :: timestep
              INTEGER,INTENT(IN)           :: number_of_timesteps
              INTEGER,INTENT(IN)           :: timesteps_between_outputs
              REAL,   INTENT(IN)           :: domain_size_x
              REAL,   INTENT(IN)           :: domain_size_y
              REAL,   INTENT(IN)           :: domain_size_z
              REAL,   INTENT(IN)           :: mass_maximum
              REAL,   INTENT(IN)           :: time_interval
              INTEGER,INTENT(IN)           :: random_seed_value

              INTEGER,PARAMETER :: output_unit = 11
              CHARACTER(LEN=default_string_length) :: filename
              INTEGER :: filename_length
              INTEGER :: index

              filename_length = LEN_TRIM(base_filename) + 1 + 8 + 1 + 3

              IF (LEN_TRIM(filename) == 0) THEN

                  WRITE (0,*)  &
     &                "ERROR: can't allocate the filename string"
                  WRITE (0,*) "  from an empty base filename"
                  WRITE (0,*) "  in ParticleArray_output"

                  CALL die_cleanly()

              END IF !! (LEN_TRIM(filename) == 0)

              WRITE(filename, "(A,A,I8.8,A)")  &
     &            TRIM(base_filename), "_", timestep, ".txt"

              OPEN(FILE=TRIM(filename), UNIT=output_unit,  &
     &             STATUS="UNKNOWN")

              WRITE (UNIT=output_unit,                              &
     &           FMT="(I4,I4,I4,G16.5,G16.5,G16.5,G16.5,G16.5,I8)") &
     &            number_of_particles, number_of_timesteps,         &
     &            timesteps_between_outputs,                        &
     &            domain_size_x, domain_size_y, domain_size_z,      &
     &            mass_maximum, time_interval,                      &
     &            random_seed_value

              WRITE(output_unit, *) timestep

              DO index = 1, number_of_particles

                  CALL Particle_output(output_unit,  &
     &                 this_particle_array(:,index))

              END DO !! index

              CLOSE(output_unit)

          END SUBROUTINE ParticleArray_output

      END MODULE nbody_particle

